use std::io;
use std::os::fd::RawFd;

use libc::epoll_event;

use crate::syscall;

const EVENTS_CAPACITY: usize = 1024;
const WAIT_MAX_EVENTS: i32 = 1024;
const WAIT_TIMEOUT: i32 = 1000;

#[repr(i32)]
#[derive(Copy, Clone)]
pub enum EpollFlags {
    // Events
    In = libc::EPOLLIN,
    Err = libc::EPOLLERR,
    Hup = libc::EPOLLHUP,
    Out = libc::EPOLLOUT,
    Pri = libc::EPOLLPRI,
    RdHup = libc::EPOLLRDHUP,

    // Flags
    Et = libc::EPOLLET,
    Exclusive = libc::EPOLLEXCLUSIVE,
    OneShot = libc::EPOLLONESHOT,
    WakeUp = libc::EPOLLWAKEUP,
}

#[repr(i32)]
pub enum EpollOp {
    Add = libc::EPOLL_CTL_ADD,
    Modify = libc::EPOLL_CTL_MOD,
    Delete = libc::EPOLL_CTL_DEL,
}

pub struct Epoll {
    fd: RawFd,
    flag: u32,
    pub events: Vec<epoll_event>,
}

impl Epoll {
    pub fn new() -> io::Result<Self> {
        Self::with_flags(&[])
    }

    pub fn with_flags(flags: &[EpollFlags]) -> io::Result<Self> {
        let flag = epoll_flags_as_u32(flags);

        match epoll_create() {
            Ok(fd) => Ok(Epoll {
                fd,
                flag,
                events: Vec::with_capacity(EVENTS_CAPACITY),
            }),
            Err(e) => Err(e),
        }
    }

    pub fn add_interest(&self, fd: RawFd, key: u64, flags: &[EpollFlags]) -> io::Result<()> {
        epoll_add_interest(self.fd, fd, self.create_epoll_event(key, flags))
    }

    pub fn modify_interest(&self, fd: RawFd, key: u64, flags: &[EpollFlags]) -> io::Result<()> {
        epoll_modify_interest(self.fd, fd, self.create_epoll_event(key, flags))
    }

    fn create_epoll_event(&self, key: u64, flags: &[EpollFlags]) -> epoll_event {
        epoll_event {
            events: self.flag | epoll_flags_as_u32(flags),
            u64: key as u64,
        }
    }

    pub fn wait(&mut self) -> io::Result<()> {
        self.events.clear();
        match epoll_wait(self.fd, &mut self.events, WAIT_MAX_EVENTS, WAIT_TIMEOUT) {
            Ok(res) => {
                // safe  as long as the kernel does nothing wrong - copied from mio
                unsafe { self.events.set_len(res) }
                Ok(())
            }
            Err(e) => Err(e)
        }
    }
}

impl EpollFlags {
    pub fn flag_match(&self, event: u32) -> bool {
        let expected_flag = *self as i32;
        event as i32 & expected_flag == expected_flag
    }
}

fn epoll_create() -> io::Result<RawFd> {
    let fd = syscall!(epoll_create1(0))?;
    if let Ok(flags) = syscall!(fcntl(fd, libc::F_GETFD)) {
        let _ = syscall!(fcntl(fd, libc::F_SETFD, flags | libc::FD_CLOEXEC));
    }

    Ok(fd)
}

fn epoll_wait(epoll_fd: RawFd, events: &mut Vec<epoll_event>, max_events: i32, timeout: i32) -> io::Result<usize> {
    match syscall!(epoll_wait(
        epoll_fd,
        events.as_mut_ptr() as *mut libc::epoll_event,
        max_events,
        timeout as libc::c_int
    )) {
        Ok(v) => Ok(v as usize),
        Err(e) => Err(e)
    }
}

fn epoll_add_interest(epoll_fd: RawFd, fd: RawFd, event: epoll_event) -> io::Result<()> {
    epoll_ctl(epoll_fd, fd, event, EpollOp::Add)
}

fn epoll_modify_interest(epoll_fd: RawFd, fd: RawFd, event: epoll_event) -> io::Result<()> {
    epoll_ctl(epoll_fd, fd, event, EpollOp::Modify)
}

fn epoll_delete_interest(epoll_fd: RawFd, fd: RawFd) -> io::Result<()> {
    let event = epoll_event { events: 0, u64: 0 };  // event will be ignored
    epoll_ctl(epoll_fd, fd, event, EpollOp::Delete)
}

fn epoll_ctl(epoll_fd: RawFd, fd: RawFd, mut event: epoll_event, op: EpollOp) -> io::Result<()> {
    syscall!(epoll_ctl(epoll_fd, op as i32, fd, &mut event))?;
    Ok(())
}

fn epoll_flags_as_u32(flags: &[EpollFlags]) -> u32 {
    let mut val = 0i32;
    flags.into_iter().for_each(|flag| val |= *flag as i32);
    val as u32
}

