use std::ffi::{CStr, CString};
use std::fs::File;
use std::io;
use std::io::Read;
use std::os::fd::{AsRawFd, FromRawFd, RawFd};

use libc::inotify_event;

use crate::buf_utils::{read_i32, read_u32};
use crate::syscall;

const INOTIFY_EVENT_BUFFER_CAPACITY: usize = 4096;
const INOTIFY_EVENT_SIZE: usize = std::mem::size_of::<inotify_event>();

#[derive(Copy, Clone, Debug)]
#[repr(u32)]
pub enum InotifyEvent {
    Access = libc::IN_ACCESS,
    Attrib = libc::IN_ATTRIB,
    CloseWrite = libc::IN_CLOSE_WRITE,
    CloseNoWrite = libc::IN_CLOSE_NOWRITE,
    Create = libc::IN_CREATE,
    Delete = libc::IN_DELETE,
    DeleteSelf = libc::IN_DELETE_SELF,
    Modify = libc::IN_MODIFY,
    MoveSelf = libc::IN_MOVE_SELF,
    MovedFrom = libc::IN_MOVED_FROM,
    MovedTo = libc::IN_MOVED_TO,
    Open = libc::IN_OPEN,
}

const IN_ALL_EVENTS: [InotifyEvent; 12] = [
    InotifyEvent::Access,
    InotifyEvent::Attrib,
    InotifyEvent::CloseWrite,
    InotifyEvent::CloseNoWrite,
    InotifyEvent::Create,
    InotifyEvent::Delete,
    InotifyEvent::DeleteSelf,
    InotifyEvent::Modify,
    InotifyEvent::MoveSelf,
    InotifyEvent::MovedFrom,
    InotifyEvent::MovedTo,
    InotifyEvent::Open,
];

pub struct Inotify {
    file: File,
    pub events: Vec<inotify_event>,
}

pub struct Watch {
    wd: RawFd,
}

impl Inotify {
    pub fn new() -> io::Result<Self> {
        let fd = inotify_init()?;
        let file = unsafe { File::from_raw_fd(fd) };

        Ok(Inotify {
            file,
            events: Vec::new(),
        })
    }

    pub fn watch(&self, path: &str, events: &[InotifyEvent]) -> io::Result<Watch> {
        let wd = inotify_add_watch(self.file.as_raw_fd(), path, events)?;
        Ok(Watch { wd })
    }

    pub fn wait(&mut self) -> io::Result<()> {
        let mut buf = [0u8; INOTIFY_EVENT_BUFFER_CAPACITY];
        let read_size = self.file.read(&mut buf)?;  // Should block

        self.events.clear();

        // Read events
        for i in 0..(read_size / INOTIFY_EVENT_SIZE) {
            let offset = INOTIFY_EVENT_SIZE * i;
            let event_buf = &buf[offset..offset + INOTIFY_EVENT_SIZE];

            self.events.push(inotify_event {
                wd: read_i32(&event_buf),
                mask: read_u32(&event_buf[4..]),
                cookie: read_u32(&event_buf[8..]),
                len: read_u32(&event_buf[12..]),
            });
        }

        Ok(())
    }
}

impl Watch {}

fn inotify_init() -> io::Result<RawFd> {
    syscall!(inotify_init())
}

fn inotify_add_watch(fd: RawFd, path: &str, events: &[InotifyEvent]) -> io::Result<RawFd> {
    let c_path = CString::new(path).unwrap();
    let mask = events_as_mask(events);

    let wd = syscall!(inotify_add_watch(fd, c_path.as_ptr(), mask))?;
    Ok(wd)
}

fn inotify_rm_watch(fd: RawFd, wd: RawFd) -> io::Result<()> {
    syscall!(inotify_rm_watch(fd, wd))?;
    Ok(())
}

fn events_as_mask(events: &[InotifyEvent]) -> u32 {
    let mut mask = 0;
    events.into_iter().for_each(|event| mask |= *event as u32);
    mask
}

pub fn mask_as_events(mask: u32) -> Vec<InotifyEvent> {
    let mut events = Vec::new();

    for event in IN_ALL_EVENTS {
        let event_u32 = event as u32;
        if mask & event_u32 == event_u32 {
            events.push(event);
        }
    }

    events
}
