const MESSAGE_METADATA_LENGTH: usize = 5;

pub trait SerializeMessage {
    fn get_message_type() -> MessageType;
    fn serialize(&self) -> Vec<u8>;
}

pub trait DeserializeMessage where Self: Sized {
    fn deserialize(buf: &[u8]) -> Self;
}

#[repr(u8)]
#[derive(Debug)]
pub enum MessageType {
    NewConnection = 1
}

impl MessageType {
    fn from(type_id: u8) -> Option<MessageType> {
        match type_id {
            1 => Some(MessageType::NewConnection),
            _ => None
        }
    }
}

#[derive(Debug)]
pub enum DeserializationError {
    InvalidDataLength(usize, usize),
    UnexpectedMessageType(u8),
}

pub fn serialize_message<T: SerializeMessage>(data: &[u8]) -> Vec<u8> {
    let mut buf = Vec::with_capacity(MESSAGE_METADATA_LENGTH);

    serialize_message_metadata::<T>(&mut buf, data.len());
    buf.extend_from_slice(data);

    buf
}

fn serialize_message_metadata<T: SerializeMessage>(buf: &mut Vec<u8>, data_length: usize) {
    let message_type = T::get_message_type() as u8;

    buf.insert(0, message_type);
    buf.insert(1, (data_length >> 24) as u8);
    buf.insert(2, (data_length >> 16) as u8);
    buf.insert(3, (data_length >> 8) as u8);
    buf.insert(4, data_length as u8);
}

pub fn read_message_data(data: &[u8]) -> Result<(MessageType, &[u8]), DeserializationError> {
    let message_type_id = data[0];
    let message_type = MessageType::from(message_type_id);
    if message_type.is_none() {
        return Err(DeserializationError::UnexpectedMessageType(message_type_id));
    }

    let data_length = deserialize_u32(&data[1..=MESSAGE_METADATA_LENGTH]) as usize;
    if data.len() < MESSAGE_METADATA_LENGTH + data_length {
        return Err(DeserializationError::InvalidDataLength(data_length, data.len()));
    }

    let data = &data[MESSAGE_METADATA_LENGTH..data_length];

    Ok((message_type.unwrap(), data))
}

pub fn deserialize_u32(data: &[u8]) -> u32 {
    let mut val = 0u32;
    for i in 0..4 {
        let offset = (3 - i) * 8;
        val += (data[i] as u32) << offset;
    }
    val
}

/// Deserialize a string at the end of a buffer, removing empty bytes at the end.
pub fn deserialize_str_at_end(buf: &[u8]) -> String {
    let mut str_end = buf.len() - 1;
    while buf[str_end] == 0 {
        str_end -= 1;
    }

    String::from_utf8_lossy(&buf[..=str_end]).into_owned()
}
