use std::fmt::{Display, Formatter};
use crate::input::events::{KvmButtonState, KvmEvent, KvmEventTrait};

pub enum KvmPointerEvent {
    MotionRelative(KvmPointerMotionRelativeEvent),
    MotionAbsolute(KvmPointerMotionAbsoluteEvent),
    Button(KvmPointerButtonEvent),
    Scroll(KvmPointerScrollEvent),
}

pub struct KvmPointerButtonEvent {
    pub button: u32,
    pub state: KvmButtonState,
}

pub struct KvmPointerMotionRelativeEvent {
    pub dx: f64,
    pub dx_unaccelerated: f64,
    pub dy: f64,
    pub dy_unaccelerated: f64,
}

pub struct KvmPointerMotionAbsoluteEvent {
    pub x: f64,
    pub x_mm: f64,
    pub y: f64,
    pub y_mm: f64,
}

pub struct KvmPointerScrollEvent {
    pub horizontal_scroll_value: f64,
    pub vertical_scroll_value: f64,
}

pub trait KvmPointerEventTrait {
    fn into_kvm_pointer_event(self) -> KvmPointerEvent;
}

impl KvmEventTrait for KvmPointerEvent {
    fn into_kvm_event(self) -> KvmEvent {
        KvmEvent::Pointer(self)
    }
}

macro_rules! kvm_event_traits {
    ($e:ident, $s:ident) => {
        impl KvmEventTrait for $s {
            fn into_kvm_event(self) -> KvmEvent {
                self.into_kvm_pointer_event().into_kvm_event()
            }
        }

        impl KvmPointerEventTrait for $s {
            fn into_kvm_pointer_event(self) -> KvmPointerEvent {
                KvmPointerEvent::$e(self)
            }
        }
    }
}

kvm_event_traits!(MotionRelative, KvmPointerMotionRelativeEvent);
kvm_event_traits!(MotionAbsolute, KvmPointerMotionAbsoluteEvent);
kvm_event_traits!(Button, KvmPointerButtonEvent);
kvm_event_traits!(Scroll, KvmPointerScrollEvent);

impl Display for KvmPointerEvent {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            KvmPointerEvent::Button(e) => write!(f, "Button({e})"),
            KvmPointerEvent::MotionRelative(e) => write!(f, "MotionRelative({e})"),
            KvmPointerEvent::MotionAbsolute(e) => write!(f, "MotionAbsolute({e})"),
            KvmPointerEvent::Scroll(e) => write!(f, "Scroll({e})")
        }
    }
}

impl Display for KvmPointerButtonEvent {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self.state {
            KvmButtonState::Pressed => write!(f, "Press {}", self.button),
            KvmButtonState::Released => write!(f, "Release {}", self.button)
        }
    }
}

impl Display for KvmPointerMotionRelativeEvent {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "x: {} ({}), y: {} ({})", self.dx, self.dx_unaccelerated, self.dy, self.dy_unaccelerated)
    }
}

impl Display for KvmPointerMotionAbsoluteEvent {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "x: {} ({}), y: {} ({})", self.x, self.x_mm, self.y, self.y_mm)
    }
}

impl Display for KvmPointerScrollEvent {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "h: {}, v: {}", self.horizontal_scroll_value, self.vertical_scroll_value)
    }
}
