use std::fs::{File, OpenOptions};
use std::os::fd::{AsRawFd, OwnedFd};
use std::os::unix::fs::OpenOptionsExt;
use std::path::Path;

use input::{Libinput, LibinputInterface};
use input::Event;
use libc::{O_RDONLY, O_RDWR, O_WRONLY};
use linux::epoll::{Epoll, EpollFlags};

use crate::input::events::KvmEvent;
use crate::input::mappings::{FromLibinputEvent, MappingError};

pub mod events;
mod mappings;

const UDEV_SEAT: &str = "seat0";
const CLIENT_SWITCH_KEY_CODE: u32 = 100; // Alt Right

pub struct InputSystem<S> {
    sub_system: S,
    epoll: Epoll,
}

struct Interface;

impl LibinputInterface for Interface {
    fn open_restricted(&mut self, path: &Path, flags: i32) -> Result<OwnedFd, i32> {
        OpenOptions::new()
            .custom_flags(flags)
            .read((flags & O_RDONLY == O_RDONLY) | (flags & O_RDWR == O_RDWR))
            .write((flags & O_WRONLY == O_WRONLY) | (flags & O_RDWR == O_RDWR))
            .open(path)
            .map(|file| file.into())
            .map_err(|err| err.raw_os_error().unwrap())
    }

    fn close_restricted(&mut self, fd: OwnedFd) {
        File::from(fd);
    }
}

impl InputSystem<Libinput> {
    pub fn new() -> Result<Self, LibinputSubsystemError> {
        let mut libinput = prepare_libinput()?;
        let epoll = create_epoll(&libinput)?;

        Ok(Self {
            sub_system: libinput,
            epoll,
        })
    }

    fn wait_for_events(&mut self) -> Result<(), LibinputSubsystemError> {
        match self.epoll.wait() {
            Ok(e) => self.dispatch(),
            Err(e) => Err(LibinputSubsystemError::Polling(format!("Failed to wait for epoll: {}", e)))
        }
    }

    fn dispatch(&mut self) -> Result<(), LibinputSubsystemError> {
        let dispatch_result = self.sub_system.dispatch();
        if dispatch_result.is_err() {
            return Err(LibinputSubsystemError::Polling("Failed to dispatch libinput events".to_string()));
        }

        Ok(())
    }
}

impl Iterator for InputSystem<Libinput> {
    type Item = KvmEvent;

    fn next(&mut self) -> Option<Self::Item> {
        let mut next = self.sub_system.next();
        if next.is_none() {
            // Try to get more events
            match self.wait_for_events() {
                Ok(_) => {
                    next = self.sub_system.next();
                }
                Err(e) => {
                    println!("Failed to wait for events: {:?}", e);
                }
            }
        }

        match next {
            Some(e) => {
                match map_libinput_event(e) {
                    Ok(e) => Some(e),
                    Err(e) => {
                        println!("Event failed: {:?}", e);
                        None
                    }
                }
            }
            None => None
        }
    }
}

#[derive(Debug)]
pub enum LibinputSubsystemError {
    Initialization(String),
    Polling(String),
    Mapping(MappingError),
}

fn prepare_libinput() -> Result<Libinput, LibinputSubsystemError> {
    let mut input = Libinput::new_with_udev(Interface);
    match input.udev_assign_seat(UDEV_SEAT) {
        Ok(_) => Ok(input),
        Err(_) => Err(LibinputSubsystemError::Initialization("Failed to assign a seat to the libinput context".to_string()))
    }
}

fn create_epoll(input: &Libinput) -> Result<Epoll, LibinputSubsystemError> {
    match Epoll::with_flags(&[EpollFlags::In]) {
        Ok(epoll) => {
            let interest = epoll.add_interest(input.as_raw_fd(), 0u64, &[]);
            match interest {
                Ok(_) => Ok(epoll),
                Err(e) => Err(LibinputSubsystemError::Initialization(format!("Failed to add epoll interest: {}", e)))
            }
        }
        Err(e) => Err(LibinputSubsystemError::Initialization(format!("Failed to create poll context: {}", e)))
    }
}

fn map_libinput_event(event: Event) -> Result<KvmEvent, LibinputSubsystemError> {
    match KvmEvent::from_libinput(event) {
        Err(error) => Err(LibinputSubsystemError::Mapping(error)),
        Ok(mapped_event) => Ok(mapped_event)
    }
}

fn intercept_client_switch_event(event: &KvmEvent) -> bool {
    match event {
        KvmEvent::Keyboard(event) if event.key_code == CLIENT_SWITCH_KEY_CODE => {
            // Do stuff
            println!("Right Alt key event");
            false
        }
        _ => true
    }
}
