use std::cmp::Ordering;
use std::collections::VecDeque;

const QUEUE_CAPACITY: usize = 1000;

pub struct MultiIndexQueue<T> {
    iterator: Box<dyn Iterator<Item=T>>,
    items: VecDeque<T>,
    offset: usize,
}

#[derive(Eq)]
pub struct QueueIndex {
    index: usize,
}

impl<T> MultiIndexQueue<T> {
    pub fn new(iterator: Box<dyn Iterator<Item=T>>) -> Self {
        Self {
            iterator,
            items: VecDeque::with_capacity(QUEUE_CAPACITY),
            offset: 0,
        }
    }

    pub fn create_index(&self) -> QueueIndex {
        QueueIndex { index: self.offset }
    }

    pub fn next(&mut self, index: &mut QueueIndex) -> Option<&T> {
        if self.items.is_empty() {
            self.fill_queue();
        }

        let real_index = index.index - self.offset;
        if real_index > QUEUE_CAPACITY {
            let next = self.fill_next();
            index.index += 1;
            return next;
        }

        let real_index = index.index - self.offset;
        let item = self.items.get(real_index);
        index.index += 1;

        item
    }

    fn fill_queue(&mut self) {
        let mut index = 0usize;
        let initial_queue_len = self.items.len();

        loop {
            if initial_queue_len + index >= QUEUE_CAPACITY {
                break;
            }

            if self.fill_next().is_none() {
                break;
            }

            index += 1;
        }
    }

    fn fill_next(&mut self) -> Option<&T> {
        let next = self.iterator.next();
        match next {
            Some(next) => {
                self.items.pop_front();
                self.items.push_back(next);
                self.offset += 1;
                self.items.get(QUEUE_CAPACITY)
            }
            None => None
        }
    }
}

impl PartialOrd<Self> for QueueIndex {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for QueueIndex {
    fn cmp(&self, other: &Self) -> Ordering {
        self.index.cmp(&other.index)
    }
}

impl PartialEq<Self> for QueueIndex {
    fn eq(&self, other: &Self) -> bool {
        self.index == other.index
    }
}
