use std::collections::HashMap;
use std::io;
use std::io::{Read, Write};
use std::net::{SocketAddr, TcpListener, TcpStream};
use std::os::fd::{AsRawFd, RawFd};

use linux::epoll::{Epoll, EpollFlags};

// Based on: https://www.zupzup.org/epoll-with-rust/index.html

const KEY_NEW_CONNECTION: u64 = 0;

pub trait TcpClient {
    fn new() -> Self;
    fn read(&self, buf: &[u8]) -> NextIntent;
    fn get_next_write(&mut self) -> (Vec<u8>, NextIntent);
    fn close(&self);
}

pub enum NextIntent {
    Read,
    Write,
}

pub struct TcpServer<T>
    where T: TcpClient {
    addr: SocketAddr,
    listener: TcpListener,
    listener_fd: RawFd,
    epoll: Epoll,
    key: u64,
    request_contexts: HashMap<u64, TcpContext<T>>,
}

struct TcpContext<T>
    where T: TcpClient {
    key: u64,
    stream: TcpStream,
    client: T,
}

impl<T> TcpServer<T>
    where T: TcpClient {
    pub fn new(addr: SocketAddr) -> io::Result<Self> {
        let listener = TcpListener::bind(addr)?;
        listener.set_nonblocking(true)?;

        let listener_fd = listener.as_raw_fd();

        let epoll = Epoll::with_flags(&[EpollFlags::OneShot])?;
        epoll.add_interest(listener_fd, KEY_NEW_CONNECTION, &[EpollFlags::In])?;

        Ok(TcpServer {
            addr,
            listener,
            listener_fd,
            epoll,
            key: 0,
            request_contexts: HashMap::new(),
        })
    }

    pub fn listen(&mut self) -> io::Result<()> {
        println!("Listening on {}", self.addr);

        loop {
            self.epoll.wait().expect("Failed to wait for epoll event");

            let events = &self.epoll.events.iter()
                .map(|event| (event.events, event.u64))
                .collect::<Vec<(u32, u64)>>();

            let mut to_remove = Vec::new();

            for (events, u64) in events {
                match *u64 {
                    KEY_NEW_CONNECTION => self.accept_connection()?,
                    key => {
                        if let Some(context) = self.request_contexts.get_mut(&key) {
                            if handle_event(&self.epoll, context, *events) {
                                context.client.close();
                                to_remove.push(context.key);
                            }
                        }
                    }
                }
            }

            for key_to_remove in to_remove {
                self.request_contexts.remove(&key_to_remove);
            }
        }
    }

    fn accept_connection(&mut self) -> io::Result<()> {
        match self.listener.accept() {
            Ok((stream, addr)) => {
                stream.set_nonblocking(true)?;
                println!("Accepted connection from {addr}");

                let client = T::new();
                let key = self.get_next_key();
                let fd = stream.as_raw_fd();

                let context = TcpContext { key, stream, client };

                self.epoll.add_interest(fd, key, &[EpollFlags::In, EpollFlags::RdHup])?;
                self.request_contexts.insert(key, context);
            }
            Err(e) => eprintln!("Couldn't accept: {e}")
        };

        self.epoll.modify_interest(self.listener_fd, KEY_NEW_CONNECTION, &[EpollFlags::In])
    }

    fn get_next_key(&mut self) -> u64 {
        self.key += 1;
        self.key
    }
}

fn handle_event<T>(epoll: &Epoll, context: &mut TcpContext<T>, event: u32) -> bool
    where T: TcpClient {
    match event {
        v if EpollFlags::In.flag_match(v) => {
            println!("Read");
            return handle_read_event(epoll, context);
        }
        v if EpollFlags::Out.flag_match(v) => {
            println!("Write");
            return handle_write_event(epoll, context);
        }
        v if EpollFlags::RdHup.flag_match(v) => {
            println!("Disconnect");
            return true;
        }
        v => println!("Unexpected event: {v}"),
    };

    false
}

fn handle_read_event<T>(epoll: &Epoll, context: &mut TcpContext<T>) -> bool
    where T: TcpClient {
    let mut data: Vec<u8> = Vec::new();
    let mut buf = [0u8; 2048];

    let read_length = context.stream.read(&mut buf).expect("Failed to read stream");
    if read_length == 0 {
        return true;
    }

    let next_interest = context.client.read(&buf);
    set_interest(epoll, context, &next_interest);

    false
}

fn handle_write_event<T>(epoll: &Epoll, context: &mut TcpContext<T>) -> bool
    where T: TcpClient {
    let (data, next_interest) = context.client.get_next_write();

    let data = trim_end(&data);
    context.stream.write(data).expect("Failed to write to stream");

    set_interest(epoll, context, &next_interest);
    false
}


fn set_interest<T>(epoll: &Epoll, context: &TcpContext<T>, next_intent: &NextIntent)
    where T: TcpClient {
    let event = match next_intent {
        NextIntent::Read => EpollFlags::In,
        NextIntent::Write => EpollFlags::Out,
    };

    epoll.modify_interest(
        context.stream.as_raw_fd(),
        context.key,
        &[event, EpollFlags::RdHup])
        .unwrap();
}

fn trim_end(buf: &[u8]) -> &[u8] {
    let mut end = buf.len() - 1;
    while buf[end] == 0 {
        end -= 1;
    }

    &buf[..=end]
}
